/*
 * ColorSchemes.h
 * --------------
 * Purpose: Various user-contributed color schemes
 * Notes  : (currently none)
 * Authors: OpenMPT Devs
 * The OpenMPT source code is released under the BSD license. Read LICENSE for more details.
 */


#pragma once

#include "openmpt/all/BuildSettings.hpp"


OPENMPT_NAMESPACE_BEGIN

struct ColorScheme
{
	const TCHAR *name;
	struct { ModColor id : 8; COLORREF color : 24; } colors[MAX_MODCOLORS];
};

inline constexpr ColorScheme ColorSchemes[] =
{
{ _T("Blue (FT2)"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xE0, 0xE0, 0x40) },
	{ MODCOLOR_BACKCURROW, RGB(0x70, 0x70, 0x70) },
	{ MODCOLOR_TEXTCURROW, RGB(0xF0, 0xF0, 0x50) },
	{ MODCOLOR_BACKSELECTED, RGB(0x40, 0x40, 0xA0) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x50, 0x50, 0x70) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xE0, 0xE0, 0x40) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x40, 0x40, 0x80) },
	{ MODCOLOR_NOTE, RGB(0xE0, 0xE0, 0x40) },
	{ MODCOLOR_INSTRUMENT, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x40, 0x40) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x2E, 0x2E, 0x5C) },
	{ MODCOLOR_SEPFACE, RGB(0x40, 0x40, 0x80) },
	{ MODCOLOR_SEPHILITE, RGB(0x99, 0x99, 0xCC) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x2E, 0x2E, 0x5A) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x40, 0x40) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x40, 0x40, 0xA0) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xF0, 0xF0, 0x50) },
}
},
{ _T("Green (IT2)"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x00, 0xE0, 0x00) },
	{ MODCOLOR_BACKCURROW, RGB(0x70, 0x70, 0x70) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0xE0, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x80, 0x80, 0x00) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0xE0, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x40, 0x68, 0x40) },
	{ MODCOLOR_NOTE, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_INSTRUMENT, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x40, 0x40) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x23, 0x38, 0x23) },
	{ MODCOLOR_SEPFACE, RGB(0x40, 0x68, 0x40) },
	{ MODCOLOR_SEPHILITE, RGB(0x94, 0xBC, 0x94) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x00, 0x40, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xE0, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x00) },
}
},
{ _T("Buzz"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xE1, 0xDB, 0xD0) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x3A, 0x34, 0x27) },
	{ MODCOLOR_BACKCURROW, RGB(0xC0, 0xB4, 0x9E) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xDD, 0xD7, 0xCC) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xA9, 0x99, 0x7A) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xCE, 0xC5, 0xB5) },
	{ MODCOLOR_NOTE, RGB(0x00, 0x00, 0x5B) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0x55, 0x55) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0x5E, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0x68, 0x68) },
	{ MODCOLOR_PITCH, RGB(0x62, 0x62, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0x66, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0xAC, 0xA8, 0xA1) },
	{ MODCOLOR_SEPFACE, RGB(0xD6, 0xD0, 0xC6) },
	{ MODCOLOR_SEPHILITE, RGB(0xEC, 0xE8, 0xE1) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xE1, 0xDB, 0xD0) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x30, 0xCC, 0x30) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x32, 0xCC, 0xCC) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xCC, 0x30) },
}
},
{ _T("arseniiv - my color scheme 2"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xFE, 0xF9, 0xE7) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKCURROW, RGB(0xC6, 0xD0, 0xD7) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0x00, 0x80, 0xC0) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFB, 0xF9, 0xAA) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0x71, 0x22) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xDD, 0xFF, 0x84) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xD1, 0xE4, 0xD6) },
	{ MODCOLOR_NOTE, RGB(0x00, 0x00, 0x80) },
	{ MODCOLOR_INSTRUMENT, RGB(0x6B, 0x6B, 0xB6) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0x66, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0x80, 0x80) },
	{ MODCOLOR_PITCH, RGB(0x80, 0x80, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0x9B, 0x88) },
	{ MODCOLOR_VUMETER_MED, RGB(0x20, 0x74, 0xFF) },
	{ MODCOLOR_VUMETER_HI, RGB(0x8B, 0x46, 0xAC) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0xA0, 0xA0, 0xA0) },
	{ MODCOLOR_SEPFACE, RGB(0xF0, 0xF0, 0xF0) },
	{ MODCOLOR_SEPHILITE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xED, 0xDF, 0xCD) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x48, 0x48) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xEF, 0xE9, 0xDE) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x24, 0x4A, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0x88, 0x28) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0x80, 0xC0) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFB, 0xF9, 0xAA) },
}
},
{ _T("barryvan - Tango"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x2E, 0x34, 0x36) },
	{ MODCOLOR_BACKCURROW, RGB(0xBA, 0xBD, 0xB6) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0x55, 0x57, 0x53) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0x8A, 0xE2, 0x34) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xBA, 0xBD, 0xB6) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xD3, 0xD7, 0xCF) },
	{ MODCOLOR_NOTE, RGB(0x20, 0x4A, 0x87) },
	{ MODCOLOR_INSTRUMENT, RGB(0x5C, 0x35, 0x66) },
	{ MODCOLOR_VOLUME, RGB(0x4E, 0x9A, 0x06) },
	{ MODCOLOR_PANNING, RGB(0xCE, 0x5C, 0x00) },
	{ MODCOLOR_PITCH, RGB(0xC4, 0xA0, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xA4, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0xFC, 0xE9, 0x4F) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFC, 0xAF, 0x3E) },
	{ MODCOLOR_VUMETER_HI, RGB(0xEF, 0x29, 0x29) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFC, 0xAF, 0x4B) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xEF, 0x29, 0x29) },
	{ MODCOLOR_SEPSHADOW, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_SEPFACE, RGB(0xD4, 0xD7, 0xCF) },
	{ MODCOLOR_SEPHILITE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xD4, 0xD0, 0xC8) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x72, 0x9F, 0xCF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xCF, 0xCF, 0x72) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x8A, 0xE2, 0x34) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x72, 0x9F, 0xCF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xCF, 0xCF, 0x72) },
}
},
{ _T("barryvan - TangoDark"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x2E, 0x34, 0x36) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x88, 0x8A, 0x85) },
	{ MODCOLOR_BACKCURROW, RGB(0x3A, 0x43, 0x45) },
	{ MODCOLOR_TEXTCURROW, RGB(0xBA, 0xBD, 0xB6) },
	{ MODCOLOR_BACKSELECTED, RGB(0x50, 0x5C, 0x5F) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0x8A, 0xE2, 0x34) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x3A, 0x43, 0x45) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x3A, 0x43, 0x45) },
	{ MODCOLOR_NOTE, RGB(0x72, 0x9F, 0xCF) },
	{ MODCOLOR_INSTRUMENT, RGB(0xAD, 0x7F, 0xA8) },
	{ MODCOLOR_VOLUME, RGB(0x8A, 0xE2, 0x34) },
	{ MODCOLOR_PANNING, RGB(0xFC, 0xAF, 0x3E) },
	{ MODCOLOR_PITCH, RGB(0xFC, 0xE9, 0x4F) },
	{ MODCOLOR_GLOBALS, RGB(0xE9, 0xB9, 0x6E) },
	{ MODCOLOR_VUMETER_LO, RGB(0xFC, 0xE9, 0x4F) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFC, 0xAF, 0x3E) },
	{ MODCOLOR_VUMETER_HI, RGB(0xEF, 0x29, 0x29) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFC, 0xAF, 0x3E) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xEF, 0x29, 0x29) },
	{ MODCOLOR_SEPSHADOW, RGB(0x62, 0x66, 0x67) },
	{ MODCOLOR_SEPFACE, RGB(0x91, 0x93, 0x8F) },
	{ MODCOLOR_SEPHILITE, RGB(0xA9, 0xAD, 0xAE) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x22, 0x27, 0x28) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x72, 0x9F, 0xCF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xCF, 0xCF, 0x72) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x8A, 0xE2, 0x34) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x72, 0x9F, 0xCF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xCF, 0xCF, 0x72) },
}
},
{ _T("Fasttracker 2 (Why Colors?)"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xFF, 0xFF, 0x82) },
	{ MODCOLOR_BACKCURROW, RGB(0x49, 0x75, 0x82) },
	{ MODCOLOR_TEXTCURROW, RGB(0xFF, 0xFF, 0x5F) },
	{ MODCOLOR_BACKSELECTED, RGB(0x28, 0x28, 0x28) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0xFF, 0xFF, 0x82) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x49, 0x75, 0x82) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x27, 0x41, 0x47) },
	{ MODCOLOR_NOTE, RGB(0xFF, 0xFF, 0x82) },
	{ MODCOLOR_INSTRUMENT, RGB(0xFF, 0xFF, 0x82) },
	{ MODCOLOR_VOLUME, RGB(0xFF, 0xFF, 0x82) },
	{ MODCOLOR_PANNING, RGB(0xFF, 0xFF, 0x82) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x82) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0xFF, 0x82) },
	{ MODCOLOR_VUMETER_LO, RGB(0x4B, 0x7C, 0x89) },
	{ MODCOLOR_VUMETER_MED, RGB(0x5F, 0x98, 0xA7) },
	{ MODCOLOR_VUMETER_HI, RGB(0xEA, 0x1C, 0x78) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xEA, 0x1C, 0x78) },
	{ MODCOLOR_SEPSHADOW, RGB(0x18, 0x28, 0x2C) },
	{ MODCOLOR_SEPFACE, RGB(0x49, 0x75, 0x82) },
	{ MODCOLOR_SEPHILITE, RGB(0x8A, 0xDB, 0xF3) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x27, 0x41, 0x47) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0x27, 0x41, 0x47) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0xFF, 0x82) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0x49, 0x75, 0x82) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x8A, 0xDB, 0xF3) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x5F) },
}
},
{ _T("herodotas - darkwater"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x00, 0x49, 0x93) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x55, 0x55) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x80, 0xC0) },
	{ MODCOLOR_BACKSELECTED, RGB(0x48, 0xC4, 0xFF) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0x80, 0xC0) },
	{ MODCOLOR_SAMPLE, RGB(0x80, 0xFF, 0xFF) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x00, 0xBF, 0xBF) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x80, 0xFF, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x00, 0x00, 0x88) },
	{ MODCOLOR_NOTE, RGB(0x00, 0xA5, 0xF4) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0xA5, 0xF4) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0x80, 0xC0) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0x00, 0x93, 0x93) },
	{ MODCOLOR_GLOBALS, RGB(0x00, 0x80, 0xFF) },
	{ MODCOLOR_VUMETER_LO, RGB(0x40, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_MED, RGB(0x80, 0x80, 0xFF) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0xFF) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x80, 0xFF, 0xFF) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0xFF) },
	{ MODCOLOR_SEPSHADOW, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_SEPFACE, RGB(0x40, 0x80, 0x80) },
	{ MODCOLOR_SEPHILITE, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x00, 0x80, 0xC0) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x16, 0x2D) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0x00, 0x55, 0x7D) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x16, 0x2D) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0x00, 0x80, 0xC0) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x80, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x32, 0xCC, 0xCC) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xCC, 0x30) },
}
},
{ _T("herodotas - grey&yellow"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x23, 0x23, 0x23) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xC7, 0xC7, 0xC7) },
	{ MODCOLOR_BACKCURROW, RGB(0x88, 0x7D, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKSELECTED, RGB(0xAA, 0xAA, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0x9F) },
	{ MODCOLOR_SAMPLE, RGB(0xE4, 0xD5, 0x0C) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x86, 0x9D, 0x02) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x48, 0x48, 0x48) },
	{ MODCOLOR_NOTE, RGB(0xEF, 0xEF, 0xEF) },
	{ MODCOLOR_INSTRUMENT, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0xF1, 0xCD, 0x01) },
	{ MODCOLOR_PANNING, RGB(0x4A, 0x8E, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xD9, 0xD9, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0x64, 0xFF, 0x37) },
	{ MODCOLOR_VUMETER_LO, RGB(0x64, 0xFF, 0x37) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xD6, 0x82, 0x03) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x20, 0xFF, 0x8F) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xE0, 0x75) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_SEPFACE, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_SEPHILITE, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x40, 0x40) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x2C, 0x2C, 0x2C) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0x60, 0x60, 0x60) },
	{ MODCOLOR_BACKENV, RGB(0x28, 0x28, 0x28) },
	{ MODCOLOR_ENVELOPES, RGB(0xF2, 0xBA, 0x11) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x30, 0xCC, 0x30) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x32, 0xCC, 0xCC) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xE4, 0xD5, 0x0C) },
}
},
{ _T("herodotas - junglist"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xFF, 0xFF, 0x9F) },
	{ MODCOLOR_BACKCURROW, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xDF, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x9D, 0x00, 0x00) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x40, 0x40, 0x40) },
	{ MODCOLOR_NOTE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_INSTRUMENT, RGB(0x80, 0xFF, 0x00) },
	{ MODCOLOR_VOLUME, RGB(0xFE, 0x12, 0x18) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x80, 0xFF, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x80, 0xFF, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x80) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x4A, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x35, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x30, 0xCC, 0x30) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x32, 0xCC, 0xCC) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xCC, 0x30) },
}
},
{ _T("herodotas - mortician"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKCURROW, RGB(0x70, 0x70, 0x70) },
	{ MODCOLOR_TEXTCURROW, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_BACKSELECTED, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x50, 0x50, 0x50) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x37, 0x37, 0x37) },
	{ MODCOLOR_NOTE, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_INSTRUMENT, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_VOLUME, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_PANNING, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_MED, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_SEPHILITE, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x60, 0x60, 0x60) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0xFF) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xC8, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xCA, 0xCA, 0xCA) },
}
},
{ _T("Hot Dog Stand"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKCURROW, RGB(0xFF, 0xC0, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0x40, 0x40, 0x40) },
	{ MODCOLOR_BACKSELECTED, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xFF, 0xE8, 0x00) },
	{ MODCOLOR_NOTE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_PANNING, RGB(0xC0, 0x80, 0x00) },
	{ MODCOLOR_PITCH, RGB(0xC0, 0x40, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xFF, 0x80, 0x00) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0x80, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x60, 0x60, 0x60) },
	{ MODCOLOR_SEPFACE, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_SEPHILITE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x80, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xFF, 0xC0, 0x00) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xE8, 0x00) },
}
},
{ _T("Impulse Tracker (Camouflage)"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x45, 0x9A, 0x49) },
	{ MODCOLOR_BACKCURROW, RGB(0x10, 0x35, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0xEB, 0xEB, 0xCB) },
	{ MODCOLOR_BACKSELECTED, RGB(0x20, 0x55, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xEB, 0xEB, 0xCB) },
	{ MODCOLOR_SAMPLE, RGB(0x8E, 0x14, 0x55) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x59, 0x41, 0x3C) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x45, 0x9A, 0x49) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x59, 0x41, 0x3C) },
	{ MODCOLOR_NOTE, RGB(0x45, 0x9A, 0x49) },
	{ MODCOLOR_INSTRUMENT, RGB(0x38, 0x9E, 0x75) },
	{ MODCOLOR_VOLUME, RGB(0xA2, 0xA2, 0xA2) },
	{ MODCOLOR_PANNING, RGB(0x18, 0x75, 0x2C) },
	{ MODCOLOR_PITCH, RGB(0x45, 0x9A, 0x49) },
	{ MODCOLOR_GLOBALS, RGB(0x45, 0x9A, 0x49) },
	{ MODCOLOR_VUMETER_LO, RGB(0x25, 0x7D, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xEC, 0x99, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0x71, 0xB5) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xEC, 0x99, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0xB6, 0x96, 0x79) },
	{ MODCOLOR_SEPFACE, RGB(0xB6, 0x96, 0x79) },
	{ MODCOLOR_SEPHILITE, RGB(0xB6, 0x96, 0x79) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x35, 0x27, 0x24) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xEB, 0xEB, 0xCB) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xA2, 0xA2, 0xA2) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x45, 0x9A, 0x49) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x38, 0x9E, 0x75) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xEB, 0xEB, 0xCB) },
}
},
{ _T("jmkz - Aqua LCD"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x0F, 0x13) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x00, 0x66, 0x80) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x66, 0x80) },
	{ MODCOLOR_TEXTCURROW, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_BACKSELECTED, RGB(0x00, 0x66, 0x80) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0xCC, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0x99, 0xBF) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x4A, 0x4A, 0x4A) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0xCC, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x00, 0x33, 0x40) },
	{ MODCOLOR_NOTE, RGB(0x00, 0xCC, 0xFF) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0x99, 0xBF) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0x99, 0xBF) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xB3, 0xDF) },
	{ MODCOLOR_PITCH, RGB(0x00, 0xB3, 0xDF) },
	{ MODCOLOR_GLOBALS, RGB(0x00, 0xCC, 0xFF) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0x66, 0x80) },
	{ MODCOLOR_VUMETER_MED, RGB(0x00, 0xCC, 0xFF) },
	{ MODCOLOR_VUMETER_HI, RGB(0x60, 0xDF, 0xFF) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x01, 0xA6, 0xFE) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x60, 0xDF, 0xFF) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x00, 0x80, 0x9F) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x00, 0x4D, 0x60) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x33, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0xCC, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xCC, 0xFF) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0x99, 0xBF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0x00, 0xB3, 0xDF) },
}
},
{ _T("jmkz - BeRoTracker"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x1E, 0x2C, 0x44) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xD2, 0xF3, 0xF7) },
	{ MODCOLOR_BACKCURROW, RGB(0xDB, 0x80, 0x24) },
	{ MODCOLOR_TEXTCURROW, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKSELECTED, RGB(0xE0, 0xB3, 0x85) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x2C, 0x39, 0x43) },
	{ MODCOLOR_SAMPLE, RGB(0x48, 0xF3, 0xA7) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xB9, 0xCC, 0xD7) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x1E, 0x2C, 0x44) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x64, 0x7A, 0x91) },
	{ MODCOLOR_NOTE, RGB(0xD2, 0xF3, 0xF7) },
	{ MODCOLOR_INSTRUMENT, RGB(0xD2, 0xF3, 0xF7) },
	{ MODCOLOR_VOLUME, RGB(0xD2, 0xF3, 0xF7) },
	{ MODCOLOR_PANNING, RGB(0xD2, 0xF3, 0xF7) },
	{ MODCOLOR_PITCH, RGB(0xD2, 0xF3, 0xF7) },
	{ MODCOLOR_GLOBALS, RGB(0xD2, 0xF3, 0xF7) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0x9B, 0xF9) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x38, 0x47, 0x58) },
	{ MODCOLOR_SEPFACE, RGB(0x38, 0x47, 0x58) },
	{ MODCOLOR_SEPHILITE, RGB(0x38, 0x47, 0x58) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x4D, 0x61, 0x85) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x48, 0xF3, 0xA7) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x48, 0xF3, 0xA7) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x64, 0x7A, 0x91) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xD2, 0xF3, 0xF7) },
}
},
{ _T("jmkz - Classic"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xF4, 0xF1, 0xEC) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x54, 0x47, 0x2C) },
	{ MODCOLOR_BACKCURROW, RGB(0xFF, 0xE2, 0x9F) },
	{ MODCOLOR_TEXTCURROW, RGB(0x51, 0x46, 0x2F) },
	{ MODCOLOR_BACKSELECTED, RGB(0x8F, 0x85, 0x70) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xEF, 0xCE) },
	{ MODCOLOR_SAMPLE, RGB(0x8F, 0xEF, 0x8F) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xFF, 0xEE, 0xCE) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x54, 0x47, 0x2C) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xE8, 0xE2, 0xD9) },
	{ MODCOLOR_NOTE, RGB(0x28, 0x23, 0x17) },
	{ MODCOLOR_INSTRUMENT, RGB(0x48, 0x43, 0x37) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0x80, 0x80) },
	{ MODCOLOR_PITCH, RGB(0x80, 0x80, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x40, 0xFF, 0x40) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x40) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x40, 0x40) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_SEPFACE, RGB(0xD5, 0xD0, 0xC8) },
	{ MODCOLOR_SEPHILITE, RGB(0xEC, 0xEC, 0xEC) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xBA, 0xAF, 0x9A) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xEF, 0xD0, 0x8F) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x8F, 0xEF, 0x8F) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x32, 0xCC, 0xCC) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xEF, 0xD0, 0x8F) },
}
},
{ _T("jmkz - Custom Colors"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x20, 0x21, 0x26) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_BACKCURROW, RGB(0x52, 0x4C, 0x34) },
	{ MODCOLOR_TEXTCURROW, RGB(0xAE, 0xA3, 0x84) },
	{ MODCOLOR_BACKSELECTED, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xD8, 0xD8, 0xD8) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0xB0, 0xB0) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x51, 0x5C, 0x7B) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xAD, 0xB3, 0xCB) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x31, 0x33, 0x3E) },
	{ MODCOLOR_NOTE, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_INSTRUMENT, RGB(0x4B, 0x8E, 0xD8) },
	{ MODCOLOR_VOLUME, RGB(0x09, 0xB3, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xB0, 0xB0) },
	{ MODCOLOR_PITCH, RGB(0xC4, 0xC4, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_VUMETER_LO, RGB(0x09, 0xB3, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0x09, 0xB3, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xCE, 0x0B, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x0C, 0x0C, 0xDC) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x0C, 0x0C, 0xDC) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xCE, 0x0B, 0x0B) },
	{ MODCOLOR_SEPSHADOW, RGB(0x18, 0x1A, 0x1D) },
	{ MODCOLOR_SEPFACE, RGB(0x40, 0x41, 0x4A) },
	{ MODCOLOR_SEPHILITE, RGB(0x74, 0x77, 0x89) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x40, 0x41, 0x53) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xC4, 0xC4, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x09, 0xB3, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0xB0, 0xB0) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xEA, 0xEA, 0x00) },
}
},
{ _T("jmkz - Custom Colors (inverted)"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xDF, 0xDE, 0xE3) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x7F, 0x7F, 0x7F) },
	{ MODCOLOR_BACKCURROW, RGB(0xAD, 0xB3, 0xCB) },
	{ MODCOLOR_TEXTCURROW, RGB(0x51, 0x5C, 0x7B) },
	{ MODCOLOR_BACKSELECTED, RGB(0x7F, 0x7F, 0x7F) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x27, 0x27, 0x27) },
	{ MODCOLOR_SAMPLE, RGB(0xFF, 0x4F, 0x4F) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xAE, 0xA3, 0x84) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xCE, 0xCC, 0xC1) },
	{ MODCOLOR_NOTE, RGB(0x3F, 0x3F, 0x3F) },
	{ MODCOLOR_INSTRUMENT, RGB(0xB4, 0x71, 0x27) },
	{ MODCOLOR_VOLUME, RGB(0xF6, 0x4C, 0xFF) },
	{ MODCOLOR_PANNING, RGB(0xFF, 0x4F, 0x4F) },
	{ MODCOLOR_PITCH, RGB(0x3B, 0x3B, 0xFF) },
	{ MODCOLOR_GLOBALS, RGB(0x3F, 0x3F, 0x3F) },
	{ MODCOLOR_VUMETER_LO, RGB(0x09, 0xB3, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0x09, 0xB3, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xCE, 0x0B, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x0C, 0x0C, 0xDC) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x0C, 0x0C, 0xDC) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xCE, 0x0B, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x8B, 0x88, 0x76) },
	{ MODCOLOR_SEPFACE, RGB(0xBF, 0xBE, 0xB5) },
	{ MODCOLOR_SEPHILITE, RGB(0xE7, 0xE5, 0xE2) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xBF, 0xBE, 0xAC) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x3B, 0x3B, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x30, 0xCC, 0x30) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x32, 0xCC, 0xCC) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xCC, 0x30) },
}
},
{ _T("jmkz - Dark Pastels"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x5F, 0x5F, 0x5F) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x31, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0x9F, 0xFF, 0x9F) },
	{ MODCOLOR_BACKSELECTED, RGB(0x00, 0x00, 0x31) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x9F, 0x9F, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0x40, 0x80, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x31, 0x00, 0x00) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0x9F, 0x9F) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x17, 0x17, 0x17) },
	{ MODCOLOR_NOTE, RGB(0xDF, 0xDF, 0xDF) },
	{ MODCOLOR_INSTRUMENT, RGB(0x80, 0xBF, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0xBF, 0xFF, 0x80) },
	{ MODCOLOR_PANNING, RGB(0x80, 0xFF, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xDF, 0x80) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x80, 0xFF) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0x80, 0x80, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0x00, 0x80) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x80, 0x80, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_SEPFACE, RGB(0x3F, 0x3F, 0x3F) },
	{ MODCOLOR_SEPHILITE, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x4F, 0x4F, 0x4F) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x40, 0x40) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x80, 0x60, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x9F, 0xFF, 0x9F) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x9F, 0x9F, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xDF, 0x80) },
}
},
{ _T("jmkz - Deep Sea"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x20, 0x40) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xAF, 0xAF, 0xAF) },
	{ MODCOLOR_BACKCURROW, RGB(0x60, 0x90, 0x30) },
	{ MODCOLOR_TEXTCURROW, RGB(0xDF, 0xEF, 0xCF) },
	{ MODCOLOR_BACKSELECTED, RGB(0x90, 0x30, 0x30) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xDF, 0xDF, 0xDF) },
	{ MODCOLOR_SAMPLE, RGB(0x9F, 0xFF, 0x40) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x30, 0x60, 0x90) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xDF, 0xDF, 0xDF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x33, 0x3F, 0x4A) },
	{ MODCOLOR_NOTE, RGB(0xED, 0xEF, 0xF1) },
	{ MODCOLOR_INSTRUMENT, RGB(0x80, 0xBF, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0xC0, 0xFF, 0x80) },
	{ MODCOLOR_PANNING, RGB(0x80, 0xC0, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xE0, 0x80) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_LO, RGB(0x80, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0x80, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0x80, 0xFF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x00, 0x80, 0xFF) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x20, 0x40) },
	{ MODCOLOR_SEPFACE, RGB(0x9F, 0x9F, 0x9F) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x20, 0x40) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x00, 0x60, 0xBF) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0xCF, 0x40) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0x80, 0xC0, 0xFF) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xC0, 0xFF, 0x80) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x80, 0xBF, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xE0, 0x80) },
}
},
{ _T("jmkz - Desert Mint"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x2C, 0x32, 0x2C) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_BACKCURROW, RGB(0x70, 0x70, 0x70) },
	{ MODCOLOR_TEXTCURROW, RGB(0xF0, 0xF0, 0x50) },
	{ MODCOLOR_BACKSELECTED, RGB(0x70, 0x96, 0x70) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_SAMPLE, RGB(0x53, 0x9A, 0x4B) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x4D, 0x4D, 0x4D) },
	{ MODCOLOR_NOTE, RGB(0x75, 0xCE, 0x40) },
	{ MODCOLOR_INSTRUMENT, RGB(0xE0, 0xDE, 0x83) },
	{ MODCOLOR_VOLUME, RGB(0x9A, 0xE0, 0x8F) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xB0, 0xB0) },
	{ MODCOLOR_PITCH, RGB(0xD2, 0xD2, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x55, 0x55) },
	{ MODCOLOR_VUMETER_LO, RGB(0xB4, 0xCC, 0xB0) },
	{ MODCOLOR_VUMETER_MED, RGB(0x65, 0xB4, 0x52) },
	{ MODCOLOR_VUMETER_HI, RGB(0x10, 0x80, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xB0, 0xB4, 0xCC) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x52, 0x65, 0xB4) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x00, 0x10, 0x80) },
	{ MODCOLOR_SEPSHADOW, RGB(0x60, 0x60, 0x60) },
	{ MODCOLOR_SEPFACE, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_SEPHILITE, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x66, 0x8A, 0x66) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0x9B, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xF0, 0xF0, 0x50) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x75, 0xCE, 0x40) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0xB0, 0xB0) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xF0, 0xF0, 0x50) },
}
},
{ _T("jmkz - EL Glow"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x0C, 0x0C, 0x0C) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_BACKCURROW, RGB(0x5F, 0x5F, 0x5F) },
	{ MODCOLOR_TEXTCURROW, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_BACKSELECTED, RGB(0x2F, 0x2F, 0x2F) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x9F, 0x9F, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x3F, 0x3F, 0x3F) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xBF, 0xBF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x1F, 0x1F, 0x1F) },
	{ MODCOLOR_NOTE, RGB(0xEE, 0xEE, 0xEE) },
	{ MODCOLOR_INSTRUMENT, RGB(0xFE, 0xDF, 0x81) },
	{ MODCOLOR_VOLUME, RGB(0xBF, 0x81, 0xFE) },
	{ MODCOLOR_PANNING, RGB(0xFF, 0x80, 0x86) },
	{ MODCOLOR_PITCH, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_GLOBALS, RGB(0x80, 0xFF, 0xFF) },
	{ MODCOLOR_VUMETER_LO, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x60, 0x80, 0xFF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_SEPSHADOW, RGB(0x0F, 0x0F, 0x0F) },
	{ MODCOLOR_SEPFACE, RGB(0x3F, 0x3F, 0x3F) },
	{ MODCOLOR_SEPHILITE, RGB(0x0F, 0x0F, 0x0F) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x5F, 0x5F, 0x5F) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x80, 0x80, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFE, 0xDF, 0x81) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x9F, 0x9F, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFE, 0xDF, 0x81) },
}
},
{ _T("jmkz - Engrayed"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x2C, 0x2C, 0x2C) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_BACKCURROW, RGB(0x50, 0x50, 0x50) },
	{ MODCOLOR_TEXTCURROW, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_BACKSELECTED, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0xC7, 0xC7, 0xC7) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x27, 0x27, 0x27) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x3F, 0x3F, 0x3F) },
	{ MODCOLOR_NOTE, RGB(0xD3, 0xD3, 0xD3) },
	{ MODCOLOR_INSTRUMENT, RGB(0x9E, 0x9E, 0x9E) },
	{ MODCOLOR_VOLUME, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_PANNING, RGB(0x9B, 0x9B, 0x9B) },
	{ MODCOLOR_PITCH, RGB(0x67, 0x67, 0x67) },
	{ MODCOLOR_GLOBALS, RGB(0xA3, 0xA3, 0xA3) },
	{ MODCOLOR_VUMETER_LO, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_VUMETER_MED, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_HI, RGB(0x60, 0x60, 0x60) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xA0, 0xA0, 0xA0) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_SEPSHADOW, RGB(0x40, 0x40, 0x40) },
	{ MODCOLOR_SEPFACE, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_SEPHILITE, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x67, 0x67, 0x67) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xAC, 0xAC, 0xAC) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xDC, 0xDC, 0xDC) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xC8, 0xC8, 0xC8) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xF0, 0xF0, 0xF0) },
}
},
{ _T("jmkz - GeneGreen"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_BACKCURROW, RGB(0xFF, 0x80, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0x40, 0x40, 0x40) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0x80, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x00, 0x37, 0x09) },
	{ MODCOLOR_NOTE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_INSTRUMENT, RGB(0xFF, 0x00, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xE0, 0x40, 0xFF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x40, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x40, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x00, 0x40, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0x80, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x00) },
}
},
{ _T("jmkz - Harmony"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xF0, 0xF8, 0xFF) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x13, 0x4D, 0x86) },
	{ MODCOLOR_BACKCURROW, RGB(0xCC, 0xE6, 0xFF) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x4D, 0x99) },
	{ MODCOLOR_BACKSELECTED, RGB(0xAA, 0xD5, 0xFF) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x13, 0x4D, 0x86) },
	{ MODCOLOR_SAMPLE, RGB(0x99, 0xFF, 0x33) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x80, 0xBF, 0xFF) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xF4, 0xF4, 0xF4) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xD9, 0xEB, 0xFF) },
	{ MODCOLOR_NOTE, RGB(0x00, 0x30, 0x60) },
	{ MODCOLOR_INSTRUMENT, RGB(0x60, 0x60, 0x60) },
	{ MODCOLOR_VOLUME, RGB(0x33, 0x99, 0xFF) },
	{ MODCOLOR_PANNING, RGB(0x99, 0x33, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0x99, 0x33) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x33, 0x99) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0xA0, 0xA0, 0xA0) },
	{ MODCOLOR_SEPFACE, RGB(0xF0, 0xF0, 0xF0) },
	{ MODCOLOR_SEPHILITE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xF0, 0xF0, 0xF0) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x33, 0xFF, 0x99) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x99, 0xFF, 0x33) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xAA, 0xD5, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xAA, 0x55) },
}
},
{ _T("jmkz - High Contrast"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_TEXTCURROW, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKSELECTED, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x80, 0x00, 0x80) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x30, 0x30, 0x30) },
	{ MODCOLOR_NOTE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_INSTRUMENT, RGB(0x80, 0xFF, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x40, 0x40) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x80, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x00) },
}
},
{ _T("jmkz - Matrix"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_BACKSELECTED, RGB(0x00, 0x60, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0xBF, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x4B, 0x4B, 0x4B) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x00, 0x40, 0x00) },
	{ MODCOLOR_NOTE, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0xBF, 0x00) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0xBF, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xDF, 0x00) },
	{ MODCOLOR_PITCH, RGB(0x00, 0xDF, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0x00, 0xBF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0xC0, 0x00) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x00, 0xE0, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x00, 0x9F, 0x00) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x00, 0x60, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0xFF) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0xBF, 0x00) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0x00, 0xE6, 0x00) },
}
},
{ _T("jmkz - Midnite Blue"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x27, 0x27, 0x78) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x00, 0x9F) },
	{ MODCOLOR_TEXTCURROW, RGB(0x7F, 0x7F, 0x7F) },
	{ MODCOLOR_BACKSELECTED, RGB(0x1F, 0x1F, 0x1F) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0x00, 0x9F) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x23, 0x23, 0x23) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x00, 0x00, 0x40) },
	{ MODCOLOR_NOTE, RGB(0xD3, 0xD3, 0xD3) },
	{ MODCOLOR_INSTRUMENT, RGB(0x9E, 0x9E, 0x9E) },
	{ MODCOLOR_VOLUME, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_PANNING, RGB(0x9B, 0x9B, 0x9B) },
	{ MODCOLOR_PITCH, RGB(0x67, 0x67, 0x67) },
	{ MODCOLOR_GLOBALS, RGB(0xA3, 0xA3, 0xA3) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0x00, 0x80) },
	{ MODCOLOR_VUMETER_MED, RGB(0x00, 0x00, 0xBF) },
	{ MODCOLOR_VUMETER_HI, RGB(0x40, 0x40, 0xFF) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x15, 0x4E, 0xE3) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x44, 0x73, 0xEE) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x69, 0x8E, 0xF1) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x2F, 0x2F, 0x2F) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x00, 0x00, 0x40) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0x00, 0xBF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0x00, 0xC8) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0x00, 0xA0) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0x62, 0x62, 0xFF) },
}
},
{ _T("jmkz - Midnite Green"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x27, 0x78, 0x27) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x9F, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0x2F, 0x2F, 0x2F) },
	{ MODCOLOR_BACKSELECTED, RGB(0x3F, 0x3F, 0x3F) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0x9F, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x5D, 0x5D, 0x5D) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x00, 0x40, 0x00) },
	{ MODCOLOR_NOTE, RGB(0xD3, 0xD3, 0xD3) },
	{ MODCOLOR_INSTRUMENT, RGB(0x9E, 0x9E, 0x9E) },
	{ MODCOLOR_VOLUME, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_PANNING, RGB(0x9B, 0x9B, 0x9B) },
	{ MODCOLOR_PITCH, RGB(0x67, 0x67, 0x67) },
	{ MODCOLOR_GLOBALS, RGB(0xA3, 0xA3, 0xA3) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0x00, 0xBF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0xC0, 0x00) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x00, 0xE0, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x2F, 0x2F, 0x2F) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x00, 0x40, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0xBF, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xC8, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0xA0, 0x00) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0x00, 0xDC, 0x00) },
}
},
{ _T("jmkz - Midnite Red"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x78, 0x27, 0x27) },
	{ MODCOLOR_BACKCURROW, RGB(0x9F, 0x00, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0x7F, 0x7F, 0x7F) },
	{ MODCOLOR_BACKSELECTED, RGB(0x2F, 0x2F, 0x2F) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0x9F, 0x00, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x2F, 0x2F, 0x2F) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x40, 0x00, 0x00) },
	{ MODCOLOR_NOTE, RGB(0xD3, 0xD3, 0xD3) },
	{ MODCOLOR_INSTRUMENT, RGB(0x9E, 0x9E, 0x9E) },
	{ MODCOLOR_VOLUME, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_PANNING, RGB(0x9B, 0x9B, 0x9B) },
	{ MODCOLOR_PITCH, RGB(0x67, 0x67, 0x67) },
	{ MODCOLOR_GLOBALS, RGB(0xA3, 0xA3, 0xA3) },
	{ MODCOLOR_VUMETER_LO, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xBF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x40, 0x40) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xE0, 0x20, 0x20) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x40, 0x40) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x2F, 0x2F, 0x2F) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x40, 0x00, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xBF, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xC8, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xB4, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0x00, 0x00) },
}
},
{ _T("jmkz - Mint"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x1F, 0x1F, 0x1F) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x9F, 0x9F, 0x9F) },
	{ MODCOLOR_BACKCURROW, RGB(0x3C, 0x3C, 0x84) },
	{ MODCOLOR_TEXTCURROW, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_BACKSELECTED, RGB(0x7F, 0x7F, 0x7F) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_SAMPLE, RGB(0xBF, 0xFF, 0x80) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x36, 0x36, 0x50) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xF0, 0xF0, 0xF0) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x2E, 0x2E, 0x2E) },
	{ MODCOLOR_NOTE, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_INSTRUMENT, RGB(0xBF, 0xFF, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_PANNING, RGB(0x80, 0xFF, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_LO, RGB(0x60, 0xFF, 0x60) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x60) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x60, 0x60) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x60, 0x60, 0xFF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x60) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x60, 0x60) },
	{ MODCOLOR_SEPSHADOW, RGB(0x2E, 0x2E, 0x2E) },
	{ MODCOLOR_SEPFACE, RGB(0x9F, 0x9F, 0x9F) },
	{ MODCOLOR_SEPHILITE, RGB(0x2E, 0x2E, 0x2E) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x9F, 0x9F, 0x9F) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xBF, 0xBF, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xBF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x80) },
}
},
{ _T("jmkz - MT2 XP Classic"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x36, 0x36, 0x2F) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xEC, 0xE9, 0xD8) },
	{ MODCOLOR_BACKCURROW, RGB(0xE0, 0xDF, 0xE3) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0x4D, 0x5D, 0x85) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0xAC, 0xA8, 0x99) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x4D, 0x61, 0x85) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x55, 0x54, 0x4A) },
	{ MODCOLOR_NOTE, RGB(0xEC, 0xE9, 0xD2) },
	{ MODCOLOR_INSTRUMENT, RGB(0xCF, 0xC8, 0x8F) },
	{ MODCOLOR_VOLUME, RGB(0xE2, 0xDE, 0xBC) },
	{ MODCOLOR_PANNING, RGB(0xE2, 0xDE, 0xBC) },
	{ MODCOLOR_PITCH, RGB(0xE2, 0xDE, 0xBC) },
	{ MODCOLOR_GLOBALS, RGB(0xEC, 0xE9, 0xD2) },
	{ MODCOLOR_VUMETER_LO, RGB(0x4D, 0x61, 0x85) },
	{ MODCOLOR_VUMETER_MED, RGB(0x85, 0x97, 0xBE) },
	{ MODCOLOR_VUMETER_HI, RGB(0xC0, 0xD2, 0xFB) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x80, 0x93, 0xB5) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x85, 0x97, 0xBE) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xC0, 0xD2, 0xFB) },
	{ MODCOLOR_SEPSHADOW, RGB(0x71, 0x6F, 0x64) },
	{ MODCOLOR_SEPFACE, RGB(0x9D, 0x9D, 0xA1) },
	{ MODCOLOR_SEPHILITE, RGB(0xF1, 0xEF, 0xE2) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x4D, 0x61, 0x85) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xAC, 0xA8, 0x99) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x46, 0xA6, 0x4E) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x4D, 0x61, 0x85) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xEC, 0xE9, 0xD2) },
}
},
{ _T("jmkz - Night Vision Red"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_BACKCURROW, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_BACKSELECTED, RGB(0x60, 0x00, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0xBF, 0x00, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x26, 0x26, 0x26) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x40, 0x00, 0x00) },
	{ MODCOLOR_NOTE, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_INSTRUMENT, RGB(0xBF, 0x00, 0x00) },
	{ MODCOLOR_VOLUME, RGB(0xBF, 0x00, 0x00) },
	{ MODCOLOR_PANNING, RGB(0xDF, 0x00, 0x00) },
	{ MODCOLOR_PITCH, RGB(0xDF, 0x00, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xBF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xE0, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x9F, 0x00, 0x00) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x60, 0x00, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xDC, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xC8, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0x42, 0x42) },
}
},
{ _T("jmkz - Oil Pastels"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x78, 0x78, 0x78) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x31, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_BACKSELECTED, RGB(0x00, 0x00, 0x40) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x80, 0x80, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0x70, 0xDF, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x40, 0x00, 0x00) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x17, 0x17, 0x17) },
	{ MODCOLOR_NOTE, RGB(0xEF, 0xEF, 0xEF) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0x80, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0x80, 0xFF, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xBF, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x00, 0xFF) },
	{ MODCOLOR_VUMETER_LO, RGB(0x80, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0x9F, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0x9F, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_SEPFACE, RGB(0x3F, 0x3F, 0x3F) },
	{ MODCOLOR_SEPHILITE, RGB(0x08, 0x08, 0x08) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x4F, 0x4F, 0x4F) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x40, 0x40) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xDF, 0xA8, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xEA, 0xEA, 0x15) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x80, 0x80, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xBF, 0x00) },
}
},
{ _T("jmkz - Old Sunset"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x93, 0x93, 0x00) },
	{ MODCOLOR_BACKCURROW, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKSELECTED, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0xDF, 0x00, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xA6, 0x00, 0x00) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x58, 0x2C, 0x2C) },
	{ MODCOLOR_NOTE, RGB(0xFF, 0x80, 0x00) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0x80, 0x80) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0x97, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0x9B, 0x9B) },
	{ MODCOLOR_PITCH, RGB(0x9D, 0x96, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xBB, 0x5E, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x20, 0xC0, 0xFF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x5B, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x97, 0x00, 0x00) },
	{ MODCOLOR_SEPHILITE, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x71, 0x00, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0x97, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0x9B, 0x9B) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xCC, 0x30) },
}
},
{ _T("jmkz - Oxygen Glow"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x0C, 0x13, 0x14) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x40, 0xAF, 0xBF) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x94, 0x9F) },
	{ MODCOLOR_TEXTCURROW, RGB(0xEF, 0xEF, 0xEF) },
	{ MODCOLOR_BACKSELECTED, RGB(0x1F, 0x5E, 0x61) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x58, 0xDC, 0xE7) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0xB1, 0xBF) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x68, 0x68, 0x68) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x40, 0xB6, 0xBF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x17, 0x27, 0x28) },
	{ MODCOLOR_NOTE, RGB(0xDF, 0xFD, 0xFF) },
	{ MODCOLOR_INSTRUMENT, RGB(0xFF, 0xD2, 0x80) },
	{ MODCOLOR_VOLUME, RGB(0xCC, 0x80, 0xFF) },
	{ MODCOLOR_PANNING, RGB(0xFF, 0x80, 0x86) },
	{ MODCOLOR_PITCH, RGB(0x80, 0xFF, 0x93) },
	{ MODCOLOR_GLOBALS, RGB(0x88, 0xEC, 0xF7) },
	{ MODCOLOR_VUMETER_LO, RGB(0x20, 0x5B, 0x60) },
	{ MODCOLOR_VUMETER_MED, RGB(0x40, 0xB6, 0xBF) },
	{ MODCOLOR_VUMETER_HI, RGB(0x80, 0xF5, 0xFF) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x7A, 0x00, 0xCC) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x99, 0x00, 0xFF) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xCC, 0x80, 0xFF) },
	{ MODCOLOR_SEPSHADOW, RGB(0x0C, 0x13, 0x14) },
	{ MODCOLOR_SEPFACE, RGB(0x1F, 0x5E, 0x61) },
	{ MODCOLOR_SEPHILITE, RGB(0x0C, 0x13, 0x14) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x00, 0x38, 0x40) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0xD9, 0x80) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xA8, 0x51) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x80, 0xFF, 0x93) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x88, 0xEC, 0xF7) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xD2, 0x80) },
}
},
{ _T("jmkz - Oxygen"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xEB, 0xF3, 0xF4) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x20, 0x59, 0x60) },
	{ MODCOLOR_BACKCURROW, RGB(0xB8, 0xE3, 0xE7) },
	{ MODCOLOR_TEXTCURROW, RGB(0x20, 0x5B, 0x60) },
	{ MODCOLOR_BACKSELECTED, RGB(0x9F, 0xD9, 0xDF) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x20, 0x5B, 0x60) },
	{ MODCOLOR_SAMPLE, RGB(0x80, 0xCE, 0xD5) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xBC, 0xC3, 0xC4) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x34, 0x49, 0x4B) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xD8, 0xE6, 0xE8) },
	{ MODCOLOR_NOTE, RGB(0x10, 0x2E, 0x30) },
	{ MODCOLOR_INSTRUMENT, RGB(0x48, 0x36, 0x17) },
	{ MODCOLOR_VOLUME, RGB(0x65, 0x30, 0x8F) },
	{ MODCOLOR_PANNING, RGB(0x8F, 0x30, 0x32) },
	{ MODCOLOR_PITCH, RGB(0x30, 0x8F, 0x3C) },
	{ MODCOLOR_GLOBALS, RGB(0x49, 0x73, 0x76) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0x9B, 0xF9) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_SEPFACE, RGB(0xC8, 0xD3, 0xD5) },
	{ MODCOLOR_SEPHILITE, RGB(0xF4, 0xF4, 0xF4) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x9A, 0xB8, 0xBA) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x40, 0xB9, 0xBF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x30, 0x8F, 0x3C) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x49, 0x73, 0x76) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x00) },
}
},
{ _T("jmkz - Pastels"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xF6, 0xF6, 0xF6) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x40, 0x40, 0x40) },
	{ MODCOLOR_BACKCURROW, RGB(0xCE, 0xFF, 0xCE) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x60, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0xCE, 0xCE, 0xFF) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0x00, 0x60) },
	{ MODCOLOR_SAMPLE, RGB(0xBF, 0xFF, 0x80) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xFF, 0xCE, 0xCE) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x60, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xE8, 0xE8, 0xE8) },
	{ MODCOLOR_NOTE, RGB(0x1F, 0x1F, 0x1F) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0x40, 0x80) },
	{ MODCOLOR_VOLUME, RGB(0x40, 0x80, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0x80, 0x80) },
	{ MODCOLOR_PITCH, RGB(0x80, 0x60, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0x80, 0x00, 0x80) },
	{ MODCOLOR_VUMETER_LO, RGB(0xBF, 0xFF, 0x80) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x80, 0xBF, 0xFF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_SEPSHADOW, RGB(0xF6, 0xF6, 0xF6) },
	{ MODCOLOR_SEPFACE, RGB(0xBF, 0xBF, 0xBF) },
	{ MODCOLOR_SEPHILITE, RGB(0xF6, 0xF6, 0xF6) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xAF, 0xAF, 0xAF) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xBF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0xDF, 0x80) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0x1A, 0x80, 0xE6) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xCE, 0xFF, 0xCE) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xCE, 0xCE, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xDF, 0x80) },
}
},
{ _T("jmkz - Plumbum"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x0C, 0x0D, 0x14) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x61, 0x65, 0x76) },
	{ MODCOLOR_BACKCURROW, RGB(0x70, 0x78, 0x8F) },
	{ MODCOLOR_TEXTCURROW, RGB(0xD8, 0xDA, 0xE7) },
	{ MODCOLOR_BACKSELECTED, RGB(0x46, 0x47, 0x59) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xD6, 0xD9, 0xE9) },
	{ MODCOLOR_SAMPLE, RGB(0x8E, 0x98, 0xB3) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x46, 0x4B, 0x59) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xCE, 0xD0, 0xE1) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x1A, 0x1C, 0x26) },
	{ MODCOLOR_NOTE, RGB(0xEF, 0xEF, 0xEF) },
	{ MODCOLOR_INSTRUMENT, RGB(0xB4, 0xBB, 0xCB) },
	{ MODCOLOR_VOLUME, RGB(0xBF, 0xD7, 0xA8) },
	{ MODCOLOR_PANNING, RGB(0xA8, 0xD7, 0xD7) },
	{ MODCOLOR_PITCH, RGB(0xD7, 0xD7, 0xA8) },
	{ MODCOLOR_GLOBALS, RGB(0xD7, 0xA8, 0xBF) },
	{ MODCOLOR_VUMETER_LO, RGB(0x34, 0x9E, 0x34) },
	{ MODCOLOR_VUMETER_MED, RGB(0xDB, 0xDB, 0x64) },
	{ MODCOLOR_VUMETER_HI, RGB(0xEF, 0x8F, 0x8F) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x34, 0x80, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xDB, 0xDB, 0x64) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xEF, 0x8F, 0x8F) },
	{ MODCOLOR_SEPSHADOW, RGB(0x10, 0x12, 0x1B) },
	{ MODCOLOR_SEPFACE, RGB(0x54, 0x55, 0x6B) },
	{ MODCOLOR_SEPHILITE, RGB(0x10, 0x12, 0x1B) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x30, 0x3C, 0x50) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xC7, 0xCB, 0xD8) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xD7, 0xD7, 0xA8) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xBF, 0xD7, 0xA8) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xA8, 0xD7, 0xD7) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xD7, 0xD7, 0xA8) },
}
},
{ _T("jmkz - Preserve (black)"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x80, 0x80, 0xFF) },
	{ MODCOLOR_BACKCURROW, RGB(0x52, 0x52, 0x74) },
	{ MODCOLOR_TEXTCURROW, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_BACKSELECTED, RGB(0x3C, 0x3C, 0x77) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x80, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0x80, 0x80, 0xC0) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x3D, 0x3D, 0x7A) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x32, 0x32, 0x47) },
	{ MODCOLOR_NOTE, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0xCE, 0xCE) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0xB7, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xC4, 0xC4) },
	{ MODCOLOR_PITCH, RGB(0xC4, 0xC4, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xED, 0x0E, 0x62) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xFF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x64, 0x64, 0x64) },
	{ MODCOLOR_SEPFACE, RGB(0x70, 0x70, 0xB8) },
	{ MODCOLOR_SEPHILITE, RGB(0x4D, 0x4D, 0x99) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x2B, 0x2B, 0x55) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x53, 0xA6, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xC4, 0xC4, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0xCE, 0xCE) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xC4, 0xC4, 0x00) },
}
},
{ _T("jmkz - Protrekkr"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x68, 0x8C, 0xAC) },
	{ MODCOLOR_BACKCURROW, RGB(0x30, 0x24, 0x34) },
	{ MODCOLOR_TEXTCURROW, RGB(0xEA, 0xF0, 0xFF) },
	{ MODCOLOR_BACKSELECTED, RGB(0x68, 0x8C, 0xAC) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0x88, 0xD0, 0xFF) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x30, 0x24, 0x34) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xDC, 0xF2, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x06, 0x1C, 0x28) },
	{ MODCOLOR_NOTE, RGB(0xFF, 0xD4, 0xB6) },
	{ MODCOLOR_INSTRUMENT, RGB(0x68, 0x8C, 0xAC) },
	{ MODCOLOR_VOLUME, RGB(0xA0, 0xBE, 0xE4) },
	{ MODCOLOR_PANNING, RGB(0xA0, 0xBE, 0xE4) },
	{ MODCOLOR_PITCH, RGB(0xA0, 0xBE, 0xE4) },
	{ MODCOLOR_GLOBALS, RGB(0xA0, 0xBE, 0xE4) },
	{ MODCOLOR_VUMETER_LO, RGB(0x10, 0xEA, 0xB0) },
	{ MODCOLOR_VUMETER_MED, RGB(0x10, 0xEA, 0xB0) },
	{ MODCOLOR_VUMETER_HI, RGB(0x10, 0xEA, 0xB0) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x28, 0x8F, 0xEC) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x28, 0x8F, 0xEC) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x28, 0x8F, 0xEC) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x88, 0xD0, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x88) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xA0, 0xBE, 0xE4) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x88, 0xD0, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x88) },
}
},
{ _T("jmkz - Pure Yellow"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x80, 0x80, 0x00) },
	{ MODCOLOR_BACKCURROW, RGB(0x80, 0x80, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_BACKSELECTED, RGB(0x60, 0x60, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0xBF, 0xBF, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x26, 0x26, 0x26) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x40, 0x40, 0x00) },
	{ MODCOLOR_NOTE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_INSTRUMENT, RGB(0xBF, 0xBF, 0x00) },
	{ MODCOLOR_VOLUME, RGB(0xBF, 0xBF, 0x00) },
	{ MODCOLOR_PANNING, RGB(0xDF, 0xDF, 0x00) },
	{ MODCOLOR_PITCH, RGB(0xDF, 0xDF, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x80, 0x80, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xBF, 0xBF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xC0, 0xC0, 0x00) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xE0, 0xE0, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPFACE, RGB(0x9F, 0x9F, 0x00) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x60, 0x60, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xCA, 0xCA, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xDF, 0xDF, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xC8, 0xC8, 0x00) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x00) },
}
},
{ _T("jmkz - Renoise Visual"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x10, 0x10, 0x15) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xAD, 0xAF, 0xB4) },
	{ MODCOLOR_BACKCURROW, RGB(0x38, 0x38, 0x38) },
	{ MODCOLOR_TEXTCURROW, RGB(0xC9, 0xC9, 0xC9) },
	{ MODCOLOR_BACKSELECTED, RGB(0x18, 0x31, 0x3C) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x66, 0x80, 0x8B) },
	{ MODCOLOR_SAMPLE, RGB(0xBF, 0xBF, 0xBF) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x2F, 0x2F, 0x2F) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xC9, 0xC9, 0xC9) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x1D, 0x20, 0x24) },
	{ MODCOLOR_NOTE, RGB(0xD1, 0xD2, 0xD5) },
	{ MODCOLOR_INSTRUMENT, RGB(0x43, 0xBD, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0xB8, 0xF6, 0xBF) },
	{ MODCOLOR_PANNING, RGB(0xE2, 0xE0, 0x93) },
	{ MODCOLOR_PITCH, RGB(0xF6, 0xB5, 0x84) },
	{ MODCOLOR_GLOBALS, RGB(0xED, 0xA6, 0xDD) },
	{ MODCOLOR_VUMETER_LO, RGB(0x07, 0xF6, 0x01) },
	{ MODCOLOR_VUMETER_MED, RGB(0xDA, 0xF6, 0x01) },
	{ MODCOLOR_VUMETER_HI, RGB(0xF6, 0x01, 0x07) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x52, 0x67, 0xFE) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xDA, 0xF6, 0x01) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xF6, 0x01, 0x07) },
	{ MODCOLOR_SEPSHADOW, RGB(0x10, 0x10, 0x15) },
	{ MODCOLOR_SEPFACE, RGB(0x70, 0x74, 0x7E) },
	{ MODCOLOR_SEPHILITE, RGB(0x10, 0x10, 0x15) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x40, 0x40, 0x54) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x64, 0xAD, 0xB1) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xC0, 0xC0, 0x54) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xB8, 0xF6, 0xBF) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x43, 0xBD, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xE2, 0xE0, 0x93) },
}
},
{ _T("jmkz - Renoise"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x10, 0x10, 0x15) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xAD, 0xAF, 0xB4) },
	{ MODCOLOR_BACKCURROW, RGB(0x38, 0x38, 0x38) },
	{ MODCOLOR_TEXTCURROW, RGB(0xC9, 0xC9, 0xC9) },
	{ MODCOLOR_BACKSELECTED, RGB(0x18, 0x31, 0x3C) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x66, 0x80, 0x8B) },
	{ MODCOLOR_SAMPLE, RGB(0xBF, 0xBF, 0xBF) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x2F, 0x2F, 0x2F) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xC9, 0xC9, 0xC9) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x1D, 0x20, 0x24) },
	{ MODCOLOR_NOTE, RGB(0xD1, 0xD2, 0xD5) },
	{ MODCOLOR_INSTRUMENT, RGB(0xD1, 0xD2, 0xD5) },
	{ MODCOLOR_VOLUME, RGB(0xB8, 0xF6, 0xBF) },
	{ MODCOLOR_PANNING, RGB(0xE2, 0xE0, 0x93) },
	{ MODCOLOR_PITCH, RGB(0xF6, 0xB5, 0x84) },
	{ MODCOLOR_GLOBALS, RGB(0xED, 0xA6, 0xDD) },
	{ MODCOLOR_VUMETER_LO, RGB(0x07, 0xF6, 0x01) },
	{ MODCOLOR_VUMETER_MED, RGB(0xDA, 0xF6, 0x01) },
	{ MODCOLOR_VUMETER_HI, RGB(0xF6, 0x01, 0x07) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x52, 0x67, 0xFE) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xDA, 0xF6, 0x01) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xF6, 0x01, 0x07) },
	{ MODCOLOR_SEPSHADOW, RGB(0x10, 0x10, 0x15) },
	{ MODCOLOR_SEPFACE, RGB(0x38, 0x3A, 0x40) },
	{ MODCOLOR_SEPHILITE, RGB(0x10, 0x10, 0x15) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x40, 0x40, 0x54) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x64, 0xAD, 0xB1) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xC0, 0xC0, 0x54) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xB8, 0xF6, 0xBF) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x43, 0xBD, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xE2, 0xE0, 0x93) },
}
},
{ _T("jmkz - Silver Luna"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xF1, 0xF0, 0xF2) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x58, 0x57, 0x68) },
	{ MODCOLOR_BACKCURROW, RGB(0xD8, 0xD8, 0xDC) },
	{ MODCOLOR_TEXTCURROW, RGB(0x58, 0x57, 0x68) },
	{ MODCOLOR_BACKSELECTED, RGB(0xB2, 0xB4, 0xBF) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLE, RGB(0xB5, 0xB4, 0xC0) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xC2, 0xC2, 0xCD) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xEA, 0xEA, 0xEE) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xE0, 0xDF, 0xE3) },
	{ MODCOLOR_NOTE, RGB(0x2C, 0x2C, 0x34) },
	{ MODCOLOR_INSTRUMENT, RGB(0x58, 0x57, 0x68) },
	{ MODCOLOR_VOLUME, RGB(0x49, 0x88, 0x40) },
	{ MODCOLOR_PANNING, RGB(0x40, 0x88, 0x7D) },
	{ MODCOLOR_PITCH, RGB(0x88, 0x7B, 0x40) },
	{ MODCOLOR_GLOBALS, RGB(0x88, 0x40, 0x4A) },
	{ MODCOLOR_VUMETER_LO, RGB(0x68, 0xE6, 0x15) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFB, 0xDC, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFC, 0x1B, 0x42) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x64, 0x64, 0xFD) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFB, 0xDC, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFC, 0x1B, 0x42) },
	{ MODCOLOR_SEPSHADOW, RGB(0x9D, 0x9D, 0xA1) },
	{ MODCOLOR_SEPFACE, RGB(0xE0, 0xDF, 0xE3) },
	{ MODCOLOR_SEPHILITE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x58, 0x57, 0x68) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xD7, 0x28, 0x47) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xDC, 0xDC, 0xE2) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xEA, 0xEA, 0x15) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x49, 0x88, 0x40) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x40, 0x88, 0x7D) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xEA, 0xEA, 0x15) },
}
},
{ _T("jmkz - Skale"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x04, 0x0E, 0x12) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x9B, 0xA4, 0xAE) },
	{ MODCOLOR_BACKCURROW, RGB(0x52, 0x60, 0x6E) },
	{ MODCOLOR_TEXTCURROW, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_BACKSELECTED, RGB(0x22, 0x68, 0x8B) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_SAMPLE, RGB(0xD6, 0xDB, 0xDE) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x36, 0x40, 0x49) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x2A, 0x35, 0x3F) },
	{ MODCOLOR_NOTE, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_INSTRUMENT, RGB(0x9B, 0xA4, 0xAE) },
	{ MODCOLOR_VOLUME, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_PANNING, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_PITCH, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_GLOBALS, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_VUMETER_LO, RGB(0x51, 0xCD, 0x32) },
	{ MODCOLOR_VUMETER_MED, RGB(0xF4, 0xC0, 0x0B) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFD, 0x0F, 0x02) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFD, 0x0F, 0x02) },
	{ MODCOLOR_SEPSHADOW, RGB(0x04, 0x0E, 0x12) },
	{ MODCOLOR_SEPFACE, RGB(0x59, 0x6B, 0x6B) },
	{ MODCOLOR_SEPHILITE, RGB(0x04, 0x0E, 0x12) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x14, 0x29, 0x2C) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xD6, 0xDB, 0xDE) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x9B, 0xA4, 0xAE) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x22, 0x68, 0x8B) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0x81, 0xC1, 0xE0) },
}
},
{ _T("jmkz - Technic"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xEF, 0xEF, 0xEA) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x52, 0x50, 0x4B) },
	{ MODCOLOR_BACKCURROW, RGB(0xC7, 0xC7, 0xC7) },
	{ MODCOLOR_TEXTCURROW, RGB(0x36, 0x36, 0x2D) },
	{ MODCOLOR_BACKSELECTED, RGB(0xE7, 0xCE, 0xC3) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x99, 0x7F, 0x74) },
	{ MODCOLOR_SAMPLE, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xD0, 0xD0, 0xD0) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x36, 0x36, 0x36) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xE2, 0xDF, 0xDB) },
	{ MODCOLOR_NOTE, RGB(0x2E, 0x2D, 0x2A) },
	{ MODCOLOR_INSTRUMENT, RGB(0x2E, 0x2D, 0x2A) },
	{ MODCOLOR_VOLUME, RGB(0x6B, 0x0E, 0x60) },
	{ MODCOLOR_PANNING, RGB(0x2C, 0x2F, 0xA0) },
	{ MODCOLOR_PITCH, RGB(0x0E, 0x70, 0xB8) },
	{ MODCOLOR_GLOBALS, RGB(0x1B, 0x85, 0x33) },
	{ MODCOLOR_VUMETER_LO, RGB(0x07, 0xF6, 0x01) },
	{ MODCOLOR_VUMETER_MED, RGB(0xDA, 0xF6, 0x01) },
	{ MODCOLOR_VUMETER_HI, RGB(0xF6, 0x01, 0x07) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x70, 0x92, 0xFE) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xDA, 0xF6, 0x01) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xF6, 0x01, 0x07) },
	{ MODCOLOR_SEPSHADOW, RGB(0xEF, 0xEF, 0xEA) },
	{ MODCOLOR_SEPFACE, RGB(0xC7, 0xC5, 0xBF) },
	{ MODCOLOR_SEPHILITE, RGB(0xEF, 0xEF, 0xEA) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xBF, 0xBF, 0xAB) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x64, 0xAC, 0xB0) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x1B, 0x85, 0x33) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x32, 0xCC, 0xCC) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x00) },
}
},
{ _T("jmkz - Tekton"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xEF, 0xEF, 0xEF) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x3F, 0x3F, 0x3F) },
	{ MODCOLOR_BACKCURROW, RGB(0xC1, 0xFF, 0xFF) },
	{ MODCOLOR_TEXTCURROW, RGB(0x40, 0x40, 0x40) },
	{ MODCOLOR_BACKSELECTED, RGB(0xFF, 0xFF, 0x9F) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x60, 0x60, 0x2A) },
	{ MODCOLOR_SAMPLE, RGB(0xBF, 0xBF, 0xBF) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xFF, 0xFF, 0xC1) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x60, 0x60, 0x60) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xE4, 0xE2, 0xDC) },
	{ MODCOLOR_NOTE, RGB(0x1F, 0x1F, 0x1F) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0x18, 0x60) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0x60, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0x60, 0x60) },
	{ MODCOLOR_PITCH, RGB(0x60, 0x60, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0x60, 0x00, 0x60) },
	{ MODCOLOR_VUMETER_LO, RGB(0xAA, 0xFF, 0xAA) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0xAE, 0xAE) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xAA, 0xAA, 0xFF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0xAE, 0xAE) },
	{ MODCOLOR_SEPSHADOW, RGB(0xEF, 0xEF, 0xEF) },
	{ MODCOLOR_SEPFACE, RGB(0x9F, 0x9F, 0x9F) },
	{ MODCOLOR_SEPHILITE, RGB(0xEF, 0xEF, 0xEF) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xBF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xE0, 0xE0, 0xE0) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x30, 0xCC, 0x30) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x2E, 0xBC, 0xBC) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x9F) },
}
},
{ _T("jmkz - Tundra"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xEC, 0xEE, 0xF4) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x2C, 0x36, 0x54) },
	{ MODCOLOR_BACKCURROW, RGB(0xCF, 0xD7, 0xF1) },
	{ MODCOLOR_TEXTCURROW, RGB(0x88, 0x94, 0xB8) },
	{ MODCOLOR_BACKSELECTED, RGB(0x70, 0x70, 0x90) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xEC, 0xEE, 0xF4) },
	{ MODCOLOR_SAMPLE, RGB(0xCA, 0x8A, 0x8D) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xBC, 0xBE, 0xC4) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xEE, 0xEF, 0xF2) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xD8, 0xDC, 0xE8) },
	{ MODCOLOR_NOTE, RGB(0x1F, 0x30, 0x61) },
	{ MODCOLOR_INSTRUMENT, RGB(0x7D, 0x67, 0x42) },
	{ MODCOLOR_VOLUME, RGB(0x65, 0x47, 0x78) },
	{ MODCOLOR_PANNING, RGB(0x84, 0x3C, 0x3E) },
	{ MODCOLOR_PITCH, RGB(0x3C, 0x84, 0x45) },
	{ MODCOLOR_GLOBALS, RGB(0x48, 0x57, 0x77) },
	{ MODCOLOR_VUMETER_LO, RGB(0x83, 0xD1, 0x83) },
	{ MODCOLOR_VUMETER_MED, RGB(0x83, 0xD1, 0x83) },
	{ MODCOLOR_VUMETER_HI, RGB(0xD1, 0x83, 0x83) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x97, 0x9F, 0xD9) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x97, 0x9F, 0xD9) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xD1, 0x83, 0x83) },
	{ MODCOLOR_SEPSHADOW, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_SEPFACE, RGB(0xC8, 0xCB, 0xD5) },
	{ MODCOLOR_SEPHILITE, RGB(0xEC, 0xEC, 0xEC) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x9A, 0xA2, 0xBA) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x96, 0xA0, 0xBF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0xCF, 0xD7, 0xF1) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x88, 0x94, 0xB8) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xE0, 0xBE, 0xC0) },
}
},
{ _T("jmkz - Xenon Glow"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x0C, 0x10, 0x14) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x41, 0x7C, 0xBF) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x43, 0x9F) },
	{ MODCOLOR_TEXTCURROW, RGB(0xEF, 0xEF, 0xEF) },
	{ MODCOLOR_BACKSELECTED, RGB(0x20, 0x39, 0x61) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x59, 0x95, 0xE7) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0x50, 0xBF) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x39, 0x39, 0x39) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x41, 0x76, 0xBF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x18, 0x1E, 0x28) },
	{ MODCOLOR_NOTE, RGB(0xDF, 0xEC, 0xFF) },
	{ MODCOLOR_INSTRUMENT, RGB(0xFF, 0x85, 0x80) },
	{ MODCOLOR_VOLUME, RGB(0x80, 0xE1, 0x8B) },
	{ MODCOLOR_PANNING, RGB(0xFF, 0xD9, 0x80) },
	{ MODCOLOR_PITCH, RGB(0xC4, 0x80, 0xFF) },
	{ MODCOLOR_GLOBALS, RGB(0x89, 0xB9, 0xF7) },
	{ MODCOLOR_VUMETER_LO, RGB(0x20, 0x3B, 0x60) },
	{ MODCOLOR_VUMETER_MED, RGB(0x41, 0x76, 0xBF) },
	{ MODCOLOR_VUMETER_HI, RGB(0x80, 0xB5, 0xFF) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x3F, 0x75, 0xBE) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x78, 0x9E, 0xD1) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xB9, 0xD7, 0xFF) },
	{ MODCOLOR_SEPSHADOW, RGB(0x0C, 0x10, 0x14) },
	{ MODCOLOR_SEPFACE, RGB(0x27, 0x46, 0x78) },
	{ MODCOLOR_SEPHILITE, RGB(0x0C, 0x10, 0x14) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x20, 0x39, 0x61) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0x80, 0x85) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x30, 0xCC, 0x30) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x59, 0x95, 0xE7) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xDF, 0xEC, 0xFF) },
}
},
{ _T("LegovitchColors"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xF2, 0xF0, 0xEC) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x3A, 0x34, 0x27) },
	{ MODCOLOR_BACKCURROW, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0x91, 0x00, 0x48) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xC6, 0x1A) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xD0, 0xDD, 0xDF) },
	{ MODCOLOR_NOTE, RGB(0x00, 0x2E, 0x5B) },
	{ MODCOLOR_INSTRUMENT, RGB(0x80, 0x80, 0x40) },
	{ MODCOLOR_VOLUME, RGB(0x1E, 0x30, 0x5B) },
	{ MODCOLOR_PANNING, RGB(0x57, 0x85, 0x2E) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0x00, 0x80) },
	{ MODCOLOR_GLOBALS, RGB(0x80, 0x00, 0xFF) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SEPFACE, RGB(0x8F, 0x8F, 0x8F) },
	{ MODCOLOR_SEPHILITE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x36, 0x57, 0x72) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x81, 0xA8, 0xD3) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xC6, 0x1A) },
}
},
{ _T("melcom - Aquamarine & Blue - Alternative)"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x14, 0x2D, 0x32) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x74, 0xA4, 0xB4) },
	{ MODCOLOR_BACKCURROW, RGB(0x38, 0x58, 0x78) },
	{ MODCOLOR_TEXTCURROW, RGB(0xBC, 0xDC, 0xF4) },
	{ MODCOLOR_BACKSELECTED, RGB(0xB4, 0xDC, 0xF4) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x28, 0x58, 0x64) },
	{ MODCOLOR_SAMPLE, RGB(0xA4, 0xCF, 0xF0) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x41, 0x66, 0x8B) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xB4, 0xDC, 0xF4) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x38, 0x58, 0x78) },
	{ MODCOLOR_NOTE, RGB(0xA4, 0xCF, 0xF0) },
	{ MODCOLOR_INSTRUMENT, RGB(0x62, 0xAC, 0xBF) },
	{ MODCOLOR_VOLUME, RGB(0xA2, 0xC1, 0xAA) },
	{ MODCOLOR_PANNING, RGB(0x8A, 0xFF, 0xA8) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x64) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_LO, RGB(0x8A, 0xFF, 0xA8) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x64) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xA4, 0xCF, 0xF0) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x64) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x80, 0x80) },
	{ MODCOLOR_SEPSHADOW, RGB(0x37, 0x79, 0x8A) },
	{ MODCOLOR_SEPFACE, RGB(0x28, 0x58, 0x64) },
	{ MODCOLOR_SEPHILITE, RGB(0x15, 0x2E, 0x35) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x28, 0x58, 0x64) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x38, 0x58, 0x78) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0x14, 0x2D, 0x32) },
	{ MODCOLOR_BACKENV, RGB(0x50, 0x7D, 0xAA) },
	{ MODCOLOR_ENVELOPES, RGB(0xA4, 0xCF, 0xF0) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0x14, 0x2D, 0x32) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x30, 0xCC, 0x30) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x32, 0xCC, 0xCC) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xCC, 0x30) },
}
},
{ _T("melcom - Brown, Olive & Green"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x2D, 0x25, 0x1E) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x98, 0xAA, 0x7D) },
	{ MODCOLOR_BACKCURROW, RGB(0x10, 0x35, 0x00) },
	{ MODCOLOR_TEXTCURROW, RGB(0xEB, 0xEB, 0xCB) },
	{ MODCOLOR_BACKSELECTED, RGB(0x20, 0x55, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xEB, 0xEB, 0xCB) },
	{ MODCOLOR_SAMPLE, RGB(0x98, 0xAA, 0x7D) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x59, 0x41, 0x3C) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x45, 0x9A, 0x49) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x43, 0x4A, 0x3E) },
	{ MODCOLOR_NOTE, RGB(0x45, 0x9A, 0x49) },
	{ MODCOLOR_INSTRUMENT, RGB(0x38, 0x9E, 0x75) },
	{ MODCOLOR_VOLUME, RGB(0xA2, 0xA2, 0xA2) },
	{ MODCOLOR_PANNING, RGB(0x8A, 0xFF, 0xA8) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x64) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_LO, RGB(0x8A, 0xFF, 0xA8) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x64) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xA4, 0xCF, 0xF0) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x64) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x80, 0x80) },
	{ MODCOLOR_SEPSHADOW, RGB(0x98, 0xAA, 0x7D) },
	{ MODCOLOR_SEPFACE, RGB(0x43, 0x4A, 0x3E) },
	{ MODCOLOR_SEPHILITE, RGB(0x2D, 0x25, 0x1E) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x35, 0x27, 0x24) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x43, 0x4A, 0x3E) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0x2D, 0x25, 0x1E) },
	{ MODCOLOR_BACKENV, RGB(0x2D, 0x25, 0x1E) },
	{ MODCOLOR_ENVELOPES, RGB(0x98, 0xAA, 0x7D) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0x43, 0x4A, 0x3E) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x30, 0xCC, 0x30) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x32, 0xCC, 0xCC) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xCC, 0x30) },
}
},
{ _T("MiDoRi - Coral Reef"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x16, 0x1D, 0x25) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x33, 0x71, 0xA4) },
	{ MODCOLOR_BACKCURROW, RGB(0x1A, 0x5F, 0x97) },
	{ MODCOLOR_TEXTCURROW, RGB(0x31, 0x2D, 0x20) },
	{ MODCOLOR_BACKSELECTED, RGB(0xF1, 0xC9, 0x8D) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x18, 0x2D, 0x43) },
	{ MODCOLOR_SAMPLE, RGB(0x1A, 0x6A, 0xA6) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x51, 0x39, 0x59) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xB4, 0x84, 0xBD) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x41, 0x52, 0x67) },
	{ MODCOLOR_NOTE, RGB(0x26, 0xD9, 0x92) },
	{ MODCOLOR_INSTRUMENT, RGB(0x9C, 0xE2, 0x8B) },
	{ MODCOLOR_VOLUME, RGB(0xEC, 0x93, 0x93) },
	{ MODCOLOR_PANNING, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_PITCH, RGB(0xC0, 0x98, 0xCB) },
	{ MODCOLOR_GLOBALS, RGB(0xD8, 0xCD, 0xA0) },
	{ MODCOLOR_VUMETER_LO, RGB(0x3A, 0xC0, 0x66) },
	{ MODCOLOR_VUMETER_MED, RGB(0xD9, 0xB4, 0x6F) },
	{ MODCOLOR_VUMETER_HI, RGB(0xDF, 0x78, 0x75) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x29, 0x7D, 0xAF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x56, 0xC5, 0xDC) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xDF, 0x78, 0x75) },
	{ MODCOLOR_SEPSHADOW, RGB(0x38, 0x3D, 0x5A) },
	{ MODCOLOR_SEPFACE, RGB(0x18, 0x1E, 0x2E) },
	{ MODCOLOR_SEPHILITE, RGB(0x32, 0x36, 0x52) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x24, 0x24, 0x2D) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x1A, 0x1C, 0x28) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_BACKENV, RGB(0x1A, 0x1C, 0x28) },
	{ MODCOLOR_ENVELOPES, RGB(0x44, 0xB9, 0x8A) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x9C, 0xE2, 0x8B) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xC0, 0x98, 0xCB) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xF1, 0xC9, 0x8D) },
}
},
{ _T("MiDoRi - DarkPlug"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x0F, 0x0F, 0x0F) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x50, 0x50, 0x50) },
	{ MODCOLOR_BACKCURROW, RGB(0x33, 0x0D, 0x30) },
	{ MODCOLOR_TEXTCURROW, RGB(0x60, 0x60, 0x60) },
	{ MODCOLOR_BACKSELECTED, RGB(0x00, 0x80, 0x80) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xE6, 0xE6, 0xE6) },
	{ MODCOLOR_SAMPLE, RGB(0x9F, 0xC7, 0x29) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x7C, 0x1A, 0x07) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x35, 0x35, 0x35) },
	{ MODCOLOR_NOTE, RGB(0xF0, 0x55, 0x44) },
	{ MODCOLOR_INSTRUMENT, RGB(0xA6, 0xA6, 0xA6) },
	{ MODCOLOR_VOLUME, RGB(0xDE, 0xC4, 0x9E) },
	{ MODCOLOR_PANNING, RGB(0x6A, 0xB9, 0xA4) },
	{ MODCOLOR_PITCH, RGB(0x4B, 0x79, 0x94) },
	{ MODCOLOR_GLOBALS, RGB(0x85, 0x97, 0x5B) },
	{ MODCOLOR_VUMETER_LO, RGB(0xA8, 0xDD, 0x6C) },
	{ MODCOLOR_VUMETER_MED, RGB(0xF8, 0xEA, 0x8B) },
	{ MODCOLOR_VUMETER_HI, RGB(0xF4, 0x8C, 0x6F) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x30, 0xAE, 0xC9) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xF8, 0xEA, 0x8B) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xF4, 0x8C, 0x6F) },
	{ MODCOLOR_SEPSHADOW, RGB(0x32, 0x22, 0x21) },
	{ MODCOLOR_SEPFACE, RGB(0x18, 0x14, 0x14) },
	{ MODCOLOR_SEPHILITE, RGB(0x35, 0x26, 0x24) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x20, 0x05, 0x00) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x12, 0x12, 0x12) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0x2B, 0x35, 0x0B) },
	{ MODCOLOR_BACKENV, RGB(0x12, 0x12, 0x12) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x6A, 0xB9, 0xA4) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x4B, 0x79, 0x94) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xAE, 0xD6, 0x38) },
}
},
{ _T("MiDoRi - DX7"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x1F, 0x1B, 0x18) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x54, 0x6D, 0x70) },
	{ MODCOLOR_BACKCURROW, RGB(0x36, 0x69, 0x62) },
	{ MODCOLOR_TEXTCURROW, RGB(0xDB, 0xD7, 0xC8) },
	{ MODCOLOR_BACKSELECTED, RGB(0xDC, 0xC5, 0xA3) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x33, 0x33, 0x33) },
	{ MODCOLOR_SAMPLE, RGB(0x3C, 0xB7, 0xA8) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x60, 0x51, 0x33) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xEC, 0xC8, 0x75) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x38, 0x31, 0x2C) },
	{ MODCOLOR_NOTE, RGB(0x6C, 0xDB, 0xCA) },
	{ MODCOLOR_INSTRUMENT, RGB(0xDF, 0xD6, 0xB7) },
	{ MODCOLOR_VOLUME, RGB(0xF7, 0xF5, 0xF0) },
	{ MODCOLOR_PANNING, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_PITCH, RGB(0x2D, 0xAC, 0x79) },
	{ MODCOLOR_GLOBALS, RGB(0x53, 0xA4, 0xB5) },
	{ MODCOLOR_VUMETER_LO, RGB(0x87, 0x5E, 0x9D) },
	{ MODCOLOR_VUMETER_MED, RGB(0xEF, 0x56, 0x56) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFA, 0xB6, 0xB4) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xBB, 0xA5, 0xC0) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xED, 0xAD, 0xAD) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0xDD, 0xDD) },
	{ MODCOLOR_SEPSHADOW, RGB(0x1B, 0x28, 0x2C) },
	{ MODCOLOR_SEPFACE, RGB(0x0D, 0x13, 0x13) },
	{ MODCOLOR_SEPHILITE, RGB(0x1F, 0x29, 0x26) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x24, 0x24, 0x2D) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xE0, 0x1F, 0x1F) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x24, 0x1F, 0x1C) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xCC, 0xBE, 0x8C) },
	{ MODCOLOR_BACKENV, RGB(0x24, 0x1F, 0x1C) },
	{ MODCOLOR_ENVELOPES, RGB(0x3C, 0xB7, 0xA8) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xB0, 0xE6, 0xE3) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x2D, 0xAC, 0x79) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x6C, 0xDB, 0xCA) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xDC, 0xC5, 0xA3) },
}
},
{ _T("MiDoRi - Midorian Mode"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x17, 0x17, 0x17) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x3E, 0x3E, 0x3E) },
	{ MODCOLOR_BACKCURROW, RGB(0x75, 0x75, 0x75) },
	{ MODCOLOR_TEXTCURROW, RGB(0x31, 0x2D, 0x20) },
	{ MODCOLOR_BACKSELECTED, RGB(0xCF, 0xF4, 0x8A) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x32, 0x27, 0x3F) },
	{ MODCOLOR_SAMPLE, RGB(0xCC, 0xE3, 0x6C) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x67, 0x60, 0x54) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xEF, 0xEF, 0xEF) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x48, 0x48, 0x48) },
	{ MODCOLOR_NOTE, RGB(0xD3, 0xF4, 0xAE) },
	{ MODCOLOR_INSTRUMENT, RGB(0xA3, 0xF0, 0x7D) },
	{ MODCOLOR_VOLUME, RGB(0x28, 0xC8, 0x31) },
	{ MODCOLOR_PANNING, RGB(0xD9, 0xC8, 0xA4) },
	{ MODCOLOR_PITCH, RGB(0xD5, 0xD7, 0xCC) },
	{ MODCOLOR_GLOBALS, RGB(0x96, 0x93, 0x72) },
	{ MODCOLOR_VUMETER_LO, RGB(0x2E, 0x47, 0x29) },
	{ MODCOLOR_VUMETER_MED, RGB(0x73, 0xB8, 0x52) },
	{ MODCOLOR_VUMETER_HI, RGB(0xDA, 0xF9, 0xCC) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x6D, 0x77, 0x4D) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x9A, 0xAE, 0x6C) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE8, 0xF0, 0xDD) },
	{ MODCOLOR_SEPSHADOW, RGB(0x29, 0x29, 0x29) },
	{ MODCOLOR_SEPFACE, RGB(0x17, 0x17, 0x17) },
	{ MODCOLOR_SEPHILITE, RGB(0x2E, 0x2E, 0x2E) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x24, 0x24, 0x2D) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xDA, 0x25, 0x25) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x2C, 0x2C, 0x2C) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0x74, 0x74, 0x74) },
	{ MODCOLOR_BACKENV, RGB(0x2C, 0x2C, 0x2C) },
	{ MODCOLOR_ENVELOPES, RGB(0xD1, 0xEB, 0x96) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0x94, 0xCB, 0x2E) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x28, 0xC8, 0x31) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xA3, 0xF0, 0x7D) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xD9, 0xC8, 0xA4) },
}
},
{ _T("MiDoRi - Strawberry"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x17, 0x1A, 0x1C) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x55, 0x60, 0x6C) },
	{ MODCOLOR_BACKCURROW, RGB(0x43, 0x4D, 0x7C) },
	{ MODCOLOR_TEXTCURROW, RGB(0x31, 0x2D, 0x20) },
	{ MODCOLOR_BACKSELECTED, RGB(0xF1, 0xC9, 0x8D) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x18, 0x2D, 0x43) },
	{ MODCOLOR_SAMPLE, RGB(0xF4, 0xAB, 0x93) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x51, 0x39, 0x59) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xB4, 0x84, 0xBD) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x36, 0x45, 0x56) },
	{ MODCOLOR_NOTE, RGB(0xEC, 0x7E, 0x6F) },
	{ MODCOLOR_INSTRUMENT, RGB(0xF1, 0xB8, 0xB8) },
	{ MODCOLOR_VOLUME, RGB(0xB1, 0xCD, 0xC7) },
	{ MODCOLOR_PANNING, RGB(0xF3, 0xFA, 0xDC) },
	{ MODCOLOR_PITCH, RGB(0xC0, 0x98, 0xCB) },
	{ MODCOLOR_GLOBALS, RGB(0x88, 0xD2, 0xB7) },
	{ MODCOLOR_VUMETER_LO, RGB(0x3A, 0xC0, 0x66) },
	{ MODCOLOR_VUMETER_MED, RGB(0xD9, 0xB4, 0x6F) },
	{ MODCOLOR_VUMETER_HI, RGB(0xDF, 0x78, 0x75) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x29, 0x7D, 0xAF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x56, 0xC5, 0xDC) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xDF, 0x78, 0x75) },
	{ MODCOLOR_SEPSHADOW, RGB(0x3F, 0x40, 0x45) },
	{ MODCOLOR_SEPFACE, RGB(0x18, 0x1E, 0x2E) },
	{ MODCOLOR_SEPHILITE, RGB(0x3F, 0x40, 0x45) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x48, 0x48, 0x5B) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFC, 0x03, 0xD7) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x36, 0x3A, 0x3F) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xB6, 0x29, 0x29) },
	{ MODCOLOR_BACKENV, RGB(0x1A, 0x1C, 0x28) },
	{ MODCOLOR_ENVELOPES, RGB(0x44, 0xB9, 0x8A) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x88, 0xD2, 0xB7) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xC0, 0x98, 0xCB) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xF3, 0xFA, 0xDC) },
}
},
{ _T("MiDoRi - Wineplume"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x1F, 0x16, 0x21) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x59, 0x44, 0x60) },
	{ MODCOLOR_BACKCURROW, RGB(0x7A, 0x27, 0x18) },
	{ MODCOLOR_TEXTCURROW, RGB(0x31, 0x2D, 0x20) },
	{ MODCOLOR_BACKSELECTED, RGB(0xF3, 0x8B, 0x8B) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x32, 0x27, 0x3F) },
	{ MODCOLOR_SAMPLE, RGB(0xED, 0x81, 0x94) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x8C, 0x38, 0x2F) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0xDC, 0x85, 0x85) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x39, 0x24, 0x3E) },
	{ MODCOLOR_NOTE, RGB(0xFC, 0xF4, 0xE4) },
	{ MODCOLOR_INSTRUMENT, RGB(0xED, 0x81, 0x94) },
	{ MODCOLOR_VOLUME, RGB(0xC8, 0xA0, 0xCB) },
	{ MODCOLOR_PANNING, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_PITCH, RGB(0xE1, 0x8C, 0xB5) },
	{ MODCOLOR_GLOBALS, RGB(0x53, 0xA4, 0xB5) },
	{ MODCOLOR_VUMETER_LO, RGB(0x87, 0x5E, 0x9D) },
	{ MODCOLOR_VUMETER_MED, RGB(0xEF, 0x56, 0x56) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFA, 0xB6, 0xB4) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0xBB, 0xA5, 0xC0) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xED, 0xAD, 0xAD) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0xDD, 0xDD) },
	{ MODCOLOR_SEPSHADOW, RGB(0x3D, 0x21, 0x37) },
	{ MODCOLOR_SEPFACE, RGB(0x37, 0x1E, 0x37) },
	{ MODCOLOR_SEPHILITE, RGB(0x39, 0x22, 0x4A) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x24, 0x24, 0x2D) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x34, 0x20, 0x39) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_BACKENV, RGB(0x34, 0x20, 0x39) },
	{ MODCOLOR_ENVELOPES, RGB(0xF9, 0xB7, 0xB7) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0x7E, 0xD6, 0xD1) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x53, 0xA4, 0xB5) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0xC8, 0xA0, 0xCB) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFC, 0xF4, 0xE4) },
}
},
{ _T("MilkyTracker"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKCURROW, RGB(0x60, 0x20, 0x40) },
	{ MODCOLOR_TEXTCURROW, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKSELECTED, RGB(0x10, 0x30, 0x60) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x17, 0x17, 0x17) },
	{ MODCOLOR_NOTE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_INSTRUMENT, RGB(0x80, 0xE0, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0xFF, 0xE0, 0x80) },
	{ MODCOLOR_PANNING, RGB(0x80, 0xFF, 0x80) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0x80, 0xE0) },
	{ MODCOLOR_GLOBALS, RGB(0x80, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x52, 0x7D, 0xAD) },
	{ MODCOLOR_SEPFACE, RGB(0x42, 0x61, 0x84) },
	{ MODCOLOR_SEPHILITE, RGB(0x21, 0x30, 0x42) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x21, 0x30, 0x42) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0x78, 0x78, 0xF0) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0x80, 0xE0, 0xFF) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x40, 0x60, 0x80) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x55, 0x80, 0xAA) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x80) },
}
},
{ _T("mp64 - Chromatic v2"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKCURROW, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_NOTE, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0xA0, 0xA0, 0xA0) },
	{ MODCOLOR_SEPFACE, RGB(0xF0, 0xF0, 0xF0) },
	{ MODCOLOR_SEPHILITE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xF0, 0xF0, 0xF0) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0x40, 0x40, 0x40) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x00) },
}
},
{ _T("mp64 - Chromatic"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_BACKCURROW, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_TEXTCURROW, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_BACKSELECTED, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0xE0, 0xE8, 0xE0) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_NOTE, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0x80, 0x00) },
	{ MODCOLOR_VOLUME, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_PANNING, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_GLOBALS, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_VUMETER_MED, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_HI, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x60, 0x60, 0x60) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x40, 0x40, 0x40) },
	{ MODCOLOR_SEPSHADOW, RGB(0xA0, 0xA0, 0xA0) },
	{ MODCOLOR_SEPFACE, RGB(0xF0, 0xF0, 0xF0) },
	{ MODCOLOR_SEPHILITE, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BLENDCOLOR, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xFF, 0x80, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xC0, 0xC0, 0xC0) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x00) },
}
},
{ _T("Oerg866 - AquaMPT"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x40) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x01, 0x54, 0xA7) },
	{ MODCOLOR_BACKCURROW, RGB(0x00, 0x40, 0x80) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0x2D, 0x45, 0x99) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xC6, 0xE1, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0x80, 0x80, 0xFF) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x00, 0x4A, 0x95) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x02, 0x21, 0x5B) },
	{ MODCOLOR_NOTE, RGB(0x97, 0xBB, 0xFF) },
	{ MODCOLOR_INSTRUMENT, RGB(0x6F, 0x88, 0xFF) },
	{ MODCOLOR_VOLUME, RGB(0x46, 0x46, 0xFF) },
	{ MODCOLOR_PANNING, RGB(0x00, 0x16, 0x2D) },
	{ MODCOLOR_PITCH, RGB(0x80, 0x80, 0xFF) },
	{ MODCOLOR_GLOBALS, RGB(0x00, 0x00, 0x66) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0x00, 0x59) },
	{ MODCOLOR_VUMETER_MED, RGB(0x00, 0x58, 0xB0) },
	{ MODCOLOR_VUMETER_HI, RGB(0x00, 0xA3, 0xF0) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x40, 0x40, 0x80) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0x60, 0x60, 0xC0) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0x80, 0x80, 0xFF) },
	{ MODCOLOR_SEPSHADOW, RGB(0x01, 0x38, 0x70) },
	{ MODCOLOR_SEPFACE, RGB(0x01, 0x3C, 0x76) },
	{ MODCOLOR_SEPHILITE, RGB(0x01, 0x60, 0xBE) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x01, 0x54, 0xA7) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x22, 0xA2, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x6F, 0x88, 0xFF) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x97, 0xBB, 0xFF) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xC6, 0xE1, 0xFF) },
}
},
{ _T("Scream Tracker 3 (Gold)"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x9A, 0x96, 0x92) },
	{ MODCOLOR_BACKCURROW, RGB(0x1A, 0x2D, 0x71) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0x14, 0x24, 0x59) },
	{ MODCOLOR_TEXTSELECTED, RGB(0x9A, 0x96, 0x92) },
	{ MODCOLOR_SAMPLE, RGB(0x00, 0xB0, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x0A, 0x11, 0x2C) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x9A, 0x96, 0x92) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x30, 0x2C, 0x28) },
	{ MODCOLOR_NOTE, RGB(0x9A, 0x96, 0x92) },
	{ MODCOLOR_INSTRUMENT, RGB(0x9A, 0x96, 0x92) },
	{ MODCOLOR_VOLUME, RGB(0x9A, 0x96, 0x92) },
	{ MODCOLOR_PANNING, RGB(0x9A, 0x96, 0x92) },
	{ MODCOLOR_PITCH, RGB(0x9A, 0x96, 0x92) },
	{ MODCOLOR_GLOBALS, RGB(0x9A, 0x96, 0x92) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0xA6, 0x92, 0x55) },
	{ MODCOLOR_SEPFACE, RGB(0xA6, 0x92, 0x55) },
	{ MODCOLOR_SEPHILITE, RGB(0xA6, 0x92, 0x55) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x69, 0x5C, 0x36) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0xA2, 0xA2, 0xA2) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x00, 0xB0, 0x00) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x22, 0x46, 0xA4) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xDF, 0x86) },
}
},
{ _T("tomaes"),
{
	{ MODCOLOR_BACKNORMAL, RGB(0x40, 0x40, 0x40) },
	{ MODCOLOR_TEXTNORMAL, RGB(0x80, 0x86, 0x93) },
	{ MODCOLOR_BACKCURROW, RGB(0x60, 0x60, 0x60) },
	{ MODCOLOR_TEXTCURROW, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKSELECTED, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_TEXTSELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_SAMPLE, RGB(0xFF, 0x00, 0x00) },
	{ MODCOLOR_BACKPLAYCURSOR, RGB(0x50, 0x50, 0x50) },
	{ MODCOLOR_TEXTPLAYCURSOR, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BACKHILIGHT, RGB(0x6D, 0x6D, 0x6D) },
	{ MODCOLOR_NOTE, RGB(0xFF, 0xA2, 0x44) },
	{ MODCOLOR_INSTRUMENT, RGB(0x00, 0xDD, 0xDD) },
	{ MODCOLOR_VOLUME, RGB(0x98, 0xC8, 0x09) },
	{ MODCOLOR_PANNING, RGB(0xFF, 0x80, 0x40) },
	{ MODCOLOR_PITCH, RGB(0xFF, 0xFF, 0x80) },
	{ MODCOLOR_GLOBALS, RGB(0xEA, 0x00, 0x4D) },
	{ MODCOLOR_VUMETER_LO, RGB(0x00, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_MED, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_VUMETER_LO_VST, RGB(0x18, 0x96, 0xE1) },
	{ MODCOLOR_VUMETER_MED_VST, RGB(0xFF, 0xC8, 0x00) },
	{ MODCOLOR_VUMETER_HI_VST, RGB(0xE1, 0x00, 0x00) },
	{ MODCOLOR_SEPSHADOW, RGB(0x80, 0x80, 0x80) },
	{ MODCOLOR_SEPFACE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SEPHILITE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_BLENDCOLOR, RGB(0x8B, 0x91, 0x9C) },
	{ MODCOLOR_DODGY_COMMANDS, RGB(0xC0, 0x00, 0x00) },
	{ MODCOLOR_BACKSAMPLE, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_SAMPLESELECTED, RGB(0xFF, 0xFF, 0xFF) },
	{ MODCOLOR_BACKENV, RGB(0x00, 0x00, 0x00) },
	{ MODCOLOR_ENVELOPES, RGB(0x00, 0x00, 0xFF) },
	{ MODCOLOR_ENVELOPE_RELEASE, RGB(0xFF, 0xFF, 0x00) },
	{ MODCOLOR_SAMPLE_LOOPMARKER, RGB(0x98, 0xC8, 0x09) },
	{ MODCOLOR_SAMPLE_SUSTAINMARKER, RGB(0x00, 0xDD, 0xDD) },
	{ MODCOLOR_SAMPLE_CUEPOINT, RGB(0xFF, 0xFF, 0x80) },
}
},
};


OPENMPT_NAMESPACE_END
